// Main program for UsbSnoop driver
// Generated by Walt Oney's driver wizard


#include "stddcls.h"
#include "driver.h"
#include "MyKdPrint.h"
#include "buffer.h"

NTSTATUS AddDevice(IN PDRIVER_OBJECT DriverObject, IN PDEVICE_OBJECT pdo);
VOID DriverUnload(IN PDRIVER_OBJECT fido);
NTSTATUS DispatchAny(IN PDEVICE_OBJECT fido, IN PIRP Irp);
NTSTATUS DispatchPower(IN PDEVICE_OBJECT fido, IN PIRP Irp);
NTSTATUS DispatchPnp(IN PDEVICE_OBJECT fido, IN PIRP Irp);
NTSTATUS DispatchWmi(IN PDEVICE_OBJECT fido, IN PIRP Irp);
NTSTATUS MyDispatchInternalIOCTL(IN PDEVICE_OBJECT fido, IN PIRP Irp);
int DumpURB(struct Buffer *b, PURB pUrb, BOOLEAN bReturnedFromHCD);

BOOLEAN        isWin98 = FALSE;
UNICODE_STRING servkey;

KDPC gDPC;

// functions to handle conversion from PipeHandle to endpint number

struct ENDPOINT_INFO 
{
    USBD_PIPE_HANDLE PipeHandle;
	unsigned char    Endpoint;
};

// the number of endpoint that can be handled is defined here.
// if you use a device with more endpoints, they won't be decoded,
// but packets will be logged anyway.

struct ENDPOINT_INFO TabEndpointInfo[MAX_ENDPOINT] =
{
	{ NULL, 0 }
};

BOOLEAN GetEndpointInfo(USBD_PIPE_HANDLE inPipeHandle, unsigned char * outEndpoint)
{
	int i;

	for (i=0;i<arraysize(TabEndpointInfo);i++)
	{
		if (TabEndpointInfo[i].PipeHandle != NULL 
			&& TabEndpointInfo[i].PipeHandle == inPipeHandle)
		{
			*outEndpoint = TabEndpointInfo[i].Endpoint;
			return TRUE;
		}
	}

	return FALSE;
}

void AddEndpointInfo(USBD_PIPE_HANDLE inPipeHandle, unsigned char inEndpoint)
{
	int i;

	// search for an existing PipeHandle
	for (i=0;i<arraysize(TabEndpointInfo);i++)
	{
		if (TabEndpointInfo[i].PipeHandle != NULL 
			&& TabEndpointInfo[i].PipeHandle == inPipeHandle)
		{
			// if found, replace the Endpoint information
			TabEndpointInfo[i].Endpoint = inEndpoint;
			return ; 
		}
	}

	// search for a free slot 
	for (i=0;i<arraysize(TabEndpointInfo);i++)
	{
		if (TabEndpointInfo[i].PipeHandle == NULL)
		{
			TabEndpointInfo[i].PipeHandle = inPipeHandle;
			TabEndpointInfo[i].Endpoint = inEndpoint;
			return ;
		}
	}

	LogPrintf("AddEndpointInfo failed!\n");
}



///////////////////////////////////////////////////////////////////////////////

void DumpStackLocation(PIO_STACK_LOCATION stack)
{
	if (stack == NULL)
		return ;

	LogPrintf("    MajorFunction=%d, MinorFunction=%d\n",
		stack->MajorFunction,stack->MinorFunction);
	LogPrintf("    DeviceObject=%p\n",stack->DeviceObject);
	LogPrintf("    CompletionRoutine=%p Context=%p\n",
		stack->CompletionRoutine,stack->Context);

}

void DumpIrp(PIRP Irp)
{
	CHAR i;

	LogPrintf("Dumping IRP %p\n",Irp);
	if (Irp==NULL)
		return ;

	LogPrintf("  Type=%d, Size=%d\n",Irp->Type,Irp->Size);
	LogPrintf("  StackCount=%d, CurrentLocation=%d\n",Irp->StackCount,
		Irp->CurrentLocation);
	for (i=0;i<Irp->StackCount;i++)
	{
		PIO_STACK_LOCATION stack = (PIO_STACK_LOCATION) (Irp+1) + i;
		LogPrintf("  [%d] MajorFunction=%d, MinorFunction=%d, DeviceObject=%p\n",
			i,stack->MajorFunction,stack->MinorFunction,stack->DeviceObject);
		LogPrintf("      Arg1=%p, Arg2=%p, Arg3=%p, Arg4=%p\n",
			stack->Parameters.Others.Argument1,stack->Parameters.Others.Argument2,
			stack->Parameters.Others.Argument3,stack->Parameters.Others.Argument4);
		LogPrintf("      CompletionRoutine=%p Context=%p\n",
			stack->CompletionRoutine,stack->Context);
	}
}

void DumpDriverObject(PDRIVER_OBJECT p)
{
	int i;

	LogPrintf("UsbSnoop - DumpDriverObject : p = %p\n",p);
	LogPrintf("  Type = %d\n",p->Type);
	LogPrintf("  Size = %d\n",p->Size);
	LogPrintf("  DeviceObject = %p\n",p->DeviceObject);
	LogPrintf("  Flags = 0x%x\n",p->Flags);
	LogPrintf("  DriverStart = %p\n",p->DriverStart);
	LogPrintf("  DriverSize = %d\n",p->DriverSize);
	LogPrintf("  DriverSection = %p\n",p->DriverSection);
	LogPrintf("  DriverExtension = %p\n",p->DriverExtension);
	LogPrintf("  DriverExtension->AddDevice = %p\n",p->DriverExtension->AddDevice);
	LogPrintf("  FastIoDispatch = %p\n",p->FastIoDispatch);
	LogPrintf("  DriverInit = %p\n",p->DriverInit);
	LogPrintf("  DriverStartIo = %p\n",p->DriverStartIo);
	LogPrintf("  DriverUnload = %p\n",p->DriverUnload);
	for (i=0;i<IRP_MJ_MAXIMUM_FUNCTION + 1;i++)
		LogPrintf("  MajorFunction[%d] = %p\n",i,p->MajorFunction[i]);
}

void DumpDeviceObject(PDEVICE_OBJECT p)
{
	PDEVICE_EXTENSION pdx;

	LogPrintf("UsbSnoop - DumpDeviceObject : p = %p\n",p);
	LogPrintf("  DriverObject = %p\n",p->DriverObject);
	LogPrintf("  NextDevice = %p\n",p->NextDevice);
	LogPrintf("  AttachedDevice = %p\n",p->AttachedDevice);
	LogPrintf("  StackSize=%d\n",p->StackSize);
	LogPrintf("  CurrentIrp = %p\n",p->CurrentIrp);
	LogPrintf("  DeviceObjectExtension = %p\n",p->DeviceObjectExtension);

	pdx = (PDEVICE_EXTENSION)p->DeviceObjectExtension;

	LogPrintf("   ->DeviceObject=%p\n",pdx->DeviceObject);
	LogPrintf("   ->LowerDeviceObject=%p\n",pdx->LowerDeviceObject);
	LogPrintf("   ->Pdo=%p\n",pdx->Pdo);
}

void DumpContext(PCONTEXT Context)
{
	LogPrintf("DumpContext : Context=%p\n",Context);
	LogPrintf("  CompletionRoutine=%p, Context=%p, Control=%x\n",
		Context->CompletionRoutine,Context->Context,Context->Control);
	LogPrintf("  pUrb=%p, uSequenceNumber=%d, Stack=%p\n",
		Context->pUrb,Context->uSequenceNumber,Context->Stack);
}

void Sleep(int nsecond)
{
	LARGE_INTEGER li;

	li.QuadPart = - nsecond * SECONDS;

	KeDelayExecutionThread(KernelMode,FALSE,&li);
}

#pragma INITCODE

/*
	DriverEntry : this function is called by Windows whenever this driver is
		loaded in memory, this is the only entry point this is accessible for
		the OS. Other entry points will be set up inside DriverEntry().

	From the Windows 2000 DDK : DriverEntry routines are called in the context
	of a system thread at IRQL PASSIVE_LEVEL.
*/

extern "C" NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
								IN PUNICODE_STRING RegistryPath)
{
	UCHAR MajorVersion, MinorVersion;
	int i;
	char msg[] = "UsbSnoop compiled on " __DATE__ " " __TIME__ " loading\n";

	LogInit();


//	DbgPrint(msg);
//	LogPrintf(msg);

	// Insist that OS support at least the WDM 1.0 (Win98 DDK)
	
	if (!IoIsWdmVersionAvailable(1, 0))
	{
		LogPrintf("UsbSnoop - Expected version of WDM (%d.%d) not available\n", 1, 0);
		return STATUS_UNSUCCESSFUL;
	}
	
	// See if we're running under Win98 or NT:
	
	isWin98 = !IoIsWdmVersionAvailable(1, 10); //DriverObject->DriverExtension->ServiceKeyName.Length == 0;
	
	MajorVersion = 0;
	MinorVersion = 0;

	while (IoIsWdmVersionAvailable(MajorVersion,MinorVersion))
		MajorVersion ++;

	MajorVersion --;

	while (IoIsWdmVersionAvailable(MajorVersion,MinorVersion))
		MinorVersion ++;
	MinorVersion --;

//	DumpDriverObject(DriverObject);

	// Save the name of the service key

	servkey.Buffer = (PWSTR) ExAllocatePool(NonPagedPool, RegistryPath->MaximumLength);
	if (servkey.Buffer == NULL)
	{
		LogPrintf("UsbSnoop - Unable to allocate %d bytes for copy of service key name\n", RegistryPath->MaximumLength);
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	servkey.MaximumLength = RegistryPath->MaximumLength;
	RtlCopyUnicodeString(&servkey, RegistryPath);

	// Initialize function pointers

	DriverObject->DriverUnload = DriverUnload;
	DriverObject->DriverExtension->AddDevice = AddDevice;
	
	for (i = 0; i < arraysize(DriverObject->MajorFunction); ++i)
		DriverObject->MajorFunction[i] = DispatchAny;

	// POWER IRP needs a special treatment
	DriverObject->MajorFunction[IRP_MJ_POWER] = DispatchPower;

	// we need a special callback to call RemoveDevice().
	// this is now done in MyDispatchPnp
	// DriverObject->MajorFunction[IRP_MJ_PNP] = DispatchPnp;

	return STATUS_SUCCESS;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

/*
	DriverUnload() is called whenever our driver is unloaded from memory.
	From Windows 2000 DDK : run at PASSIVE_LEVEL, as are any driver-created
	system threads. 
*/

VOID DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
	char msg [] = "Driver unloaded!\n";

//	LogPrintf("UsbSnoop - DriverUnload(%p) : DriverObject %p, IRQL=%d\n",
//		DriverUnload,DriverObject,KeGetCurrentIrql());

//	DumpDriverObject(DriverObject);
	RtlFreeUnicodeString(&servkey);

//	LogPrintf(msg);
	LogDone();

	/*
		let the time for the thread to stop. It's no longer needed, since we
		use an event to synchronize the destruction of the thread.
		after that, we CANNOT call LogPrintf).
	*/

	DbgPrint(msg);
}

const char * GetIrpPnpMinorFunctionName(ULONG fcn)
{
	static const char* fcnname[] =
	{
		"IRP_MN_START_DEVICE",
		"IRP_MN_QUERY_REMOVE_DEVICE",
		"IRP_MN_REMOVE_DEVICE",
		"IRP_MN_CANCEL_REMOVE_DEVICE",
		"IRP_MN_STOP_DEVICE",
		"IRP_MN_QUERY_STOP_DEVICE",
		"IRP_MN_CANCEL_STOP_DEVICE",
		"IRP_MN_QUERY_DEVICE_RELATIONS",
		"IRP_MN_QUERY_INTERFACE",
		"IRP_MN_QUERY_CAPABILITIES",
		"IRP_MN_QUERY_RESOURCES",
		"IRP_MN_QUERY_RESOURCE_REQUIREMENTS",
		"IRP_MN_QUERY_DEVICE_TEXT",
		"IRP_MN_FILTER_RESOURCE_REQUIREMENTS",
		"",
		"IRP_MN_READ_CONFIG",
		"IRP_MN_WRITE_CONFIG",
		"IRP_MN_EJECT",
		"IRP_MN_SET_LOCK",
		"IRP_MN_QUERY_ID",
		"IRP_MN_QUERY_PNP_DEVICE_STATE",
		"IRP_MN_QUERY_BUS_INFORMATION",
		"IRP_MN_DEVICE_USAGE_NOTIFICATION",
		"IRP_MN_SURPRISE_REMOVAL",
	};

	if (0<=fcn && fcn<arraysize(fcnname))
		return fcnname[fcn];

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID RemoveDevice(IN PDEVICE_OBJECT fido)
{
//	PAGED_CODE();

	// since we called this function before the original one,
	// we are sure that the lower device object is still the same

	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;

//	LogPrintf("UsbSnoop - RemoveDevice(%p) : fido=%p pdx=%p\n",RemoveDevice,fido,pdx);
	
	if (pdx->LowerDeviceObject != NULL)
	{
		// save our modified PDRIVER_OBJECT
		PDRIVER_OBJECT d = pdx->LowerDeviceObject->DriverObject;

//		LogPrintf("  fdo=%p OriginalDriverObject=%p d=%p\n",
//			pdx->LowerDeviceObject,pdx->OriginalDriverObject,d);

		// restore the driver pointer in the PDO
		pdx->LowerDeviceObject->DriverObject = pdx->OriginalDriverObject;

		IoDetachDevice(pdx->LowerDeviceObject);

		// free our modified PDRIVER_OBJECT
		ExFreePool(d);
	}

	IoDeleteDevice(fido);
}

NTSTATUS MyDispatchPnp(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG fcn = stack->MinorFunction;
	PDEVICE_OBJECT fido;
//	NTSTATUS status;
	PDEVICE_EXTENSION pdx;
	
	const char * MinorFunctionName = GetIrpPnpMinorFunctionName(fcn);
//	if (MinorFunctionName != NULL)
//		LogPrintf("UsbSnoop - MyDispatchPNP(%p) : IRP_MJ_PNP (%s)\n",
//			MyDispatchPnp,MinorFunctionName);
//	else
//		LogPrintf("UsbSnoop - MyDispatchPNP(%p) : IRP_MJ_PNP (0x%x)\n",
//			MyDispatchPnp,fcn);

	
	fido = fdo->AttachedDevice;
//	LogPrintf("fido=%p\n",fido);

	pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
//	LogPrintf("pdx=%p\n",pdx);

	// save the original function to call

	PDRIVER_DISPATCH pfn = pdx->OriginalDriverObject->MajorFunction[IRP_MJ_PNP];

	// the following lines should be executed ONLY if DispatchPNP() callback
	// is not used, since it duplicates its work.

	// They should also be executed BEFORE calling the original Dispatch function
	// since this original Dispatch function will surely destroy its DeviceObject.

	if (fcn == IRP_MN_REMOVE_DEVICE)
		RemoveDevice(fido);
	
	return pfn(fdo,Irp);	
}

/*
	Experimentally, under Windows 2000, run at IRQL DISPATCH_LEVEL,
	which is not enought for us, since ZwWriteFile() must be called
	at IRQL PASSIVE_LEVEL
*/

VOID gDeferredRoutine(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{

	KIRQL currentIrql = KeGetCurrentIrql();
	DbgPrint("gDeferredRoutine : currentIrql=%d\n",currentIrql);
}

/*
	AddDevice is called by Windows for each device we have to managed. This is
	where we get the oportunity to be added in the device object stack.

  From the Windows 2000 DDK : An AddDevice routine is called in the context of
  a system thread at IRQL PASSIVE_LEVEL.

  A device driver can register a CustomDpc routine by calling KeInitializeDpc
  after creating a device object. The driver should make this call from its
  AddDevice routine. Callers of this routine must be running at
  IRQL PASSIVE_LEVEL.
*/

NTSTATUS AddDevice(IN PDRIVER_OBJECT DriverObject, IN PDEVICE_OBJECT pdo)
{
	NTSTATUS status;
	PDEVICE_OBJECT fido, fdo;
	PDEVICE_EXTENSION pdx;
	PDRIVER_OBJECT d;

//	PAGED_CODE();

//	LogPrintf("UsbSnoop - AddDevice(%p) : DriverObject %p, pdo %p\n",
//		AddDevice,DriverObject, pdo);

//	DumpDriverObject(DriverObject);
//	DumpDeviceObject(pdo);

	// create a custom DPC
	//KeInitializeDpc(&gDPC,gDeferredRoutine,NULL);

	// Create a UsbSnoop device object to represent the hardware we're managing.
	
	status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),NULL,
		FILE_DEVICE_UNKNOWN, 0, FALSE, &fido);
	if (!NT_SUCCESS(status))
	{						// can't create device object
		LogPrintf("UsbSnoop - IoCreateDevice failed - %x\n", status);
		return status;
	}						// can't create device object

	// Benoit PAPILLAULT 13/07/2001
	// fido->DeviceExtension is a user define structure whose size is passed
	// to IoCreateDevice(). We can store whatever we want inside.

	pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
//	LogPrintf("  fido=%p pdx=%p\n",fido,pdx);
	
	// From this point forward, any error will have side effects that need to
	// be cleaned up. Using a try-finally block allows us to modify the program
	// easily without losing track of the side effects.
	
	__try
	{						// finish initialization
		IoInitializeRemoveLock(&pdx->RemoveLock, 0, 0, 255);
		pdx->DeviceObject = fido;
		pdx->Pdo = pdo;
		
		// Add our device object to the stack and propagate critical settings
		// from the immediately lower device object
		
		fdo = IoAttachDeviceToDeviceStack(fido, pdo);

		pdx->LowerDeviceObject = fdo;
		fido->Flags |= fdo->Flags & (DO_DIRECT_IO | DO_BUFFERED_IO | DO_POWER_PAGABLE | DO_POWER_INRUSH);
		fido->DeviceType = fdo->DeviceType;
		fido->Characteristics = fdo->Characteristics;
		fido->AlignmentRequirement = fdo->AlignmentRequirement;
		
		// Clear the "initializing" flag so that we can get IRPs
		
		fido->Flags &= ~DO_DEVICE_INITIALIZING;

//		DumpDriverObject(DriverObject);
//		DumpDeviceObject(fdo);
//		DumpDeviceObject(fido);
//		DumpDeviceObject(pdo);

		// we make a copy of fdo->DriverObject
		d = (PDRIVER_OBJECT)ExAllocatePool(NonPagedPool,sizeof(DRIVER_OBJECT));
		if (d != NULL)
		{
			*d = *fdo->DriverObject;

			// we make some changes to this copy
			d->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL] = MyDispatchInternalIOCTL;
			d->MajorFunction[IRP_MJ_PNP]                     = MyDispatchPnp;

			// here is the trick : we save the original DriverObject
			// and next, it points to our modified copy
			pdx->OriginalDriverObject = fdo->DriverObject;
			fdo->DriverObject = d;

//			LogPrintf("  fdo=%p OriginalDriverObject=%p d=%p\n",
//				fdo,pdx->OriginalDriverObject,d);
		}
		else
			LogPrintf("ExAllocatePool failed : not redirecting PDO->DriverObject\n");
	}						// finish initialization
	__finally
	{						// cleanup side effects
		if (!NT_SUCCESS(status))
		{					// need to cleanup
			IoDeleteDevice(fido);
		}					// need to cleanup
	}						// cleanup side effects
	
	return status;
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

NTSTATUS CompleteRequest(IN PIRP Irp, IN NTSTATUS status, IN ULONG info)
{
	Irp->IoStatus.Status = status;
	Irp->IoStatus.Information = info;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return status;
}

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE				// make no assumptions about pageability of dispatch fcns

NTSTATUS DispatchAny(IN PDEVICE_OBJECT fido, IN PIRP Irp)
{
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	NTSTATUS status;
	
	const char* irpname[] =
	{
		"IRP_MJ_CREATE",
		"IRP_MJ_CREATE_NAMED_PIPE",
		"IRP_MJ_CLOSE",
		"IRP_MJ_READ",
		"IRP_MJ_WRITE",
		"IRP_MJ_QUERY_INFORMATION",
		"IRP_MJ_SET_INFORMATION",
		"IRP_MJ_QUERY_EA",
		"IRP_MJ_SET_EA",
		"IRP_MJ_FLUSH_BUFFERS",
		"IRP_MJ_QUERY_VOLUME_INFORMATION",
		"IRP_MJ_SET_VOLUME_INFORMATION",
		"IRP_MJ_DIRECTORY_CONTROL",
		"IRP_MJ_FILE_SYSTEM_CONTROL",
		"IRP_MJ_DEVICE_CONTROL",
		"IRP_MJ_INTERNAL_DEVICE_CONTROL",
		"IRP_MJ_SHUTDOWN",
		"IRP_MJ_LOCK_CONTROL",
		"IRP_MJ_CLEANUP",
		"IRP_MJ_CREATE_MAILSLOT",
		"IRP_MJ_QUERY_SECURITY",
		"IRP_MJ_SET_SECURITY",
		"IRP_MJ_POWER",
		"IRP_MJ_SYSTEM_CONTROL",
		"IRP_MJ_DEVICE_CHANGE",
		"IRP_MJ_QUERY_QUOTA",
		"IRP_MJ_SET_QUOTA",
		"IRP_MJ_PNP",
	};
	
	UCHAR type = stack->MajorFunction;

	if (type == IRP_MJ_PNP)
	{
		ULONG fcn = stack->MinorFunction;
	
		const char * MinorFunctionName = GetIrpPnpMinorFunctionName(fcn);
//		if (MinorFunctionName != NULL)
//			LogPrintf("UsbSnoop - DispatchAny(%p) : IRP_MJ_PNP (%s)\n",
//				DispatchAny,MinorFunctionName);
//		else
//			LogPrintf("UsbSnoop - DispatchAny(%p) : IRP_MJ_PNP (0x%x)\n",
//				DispatchAny,fcn);
	}
	else
	{
//		if (type >= arraysize(irpname))
//			LogPrintf("UsbSnoop - DispatchAny(%p) : Unknown IRP, MajorFunction=0x%x\n",
//				DispatchAny,type);
//		else
//			LogPrintf("UsbSnoop - DispatchAny(%p) : %s\n",DispatchAny, irpname[type]);
	}
	
	// Pass request down without additional processing
	
	status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
	if (!NT_SUCCESS(status))
		return CompleteRequest(Irp, status, 0);
	IoSkipCurrentIrpStackLocation(Irp);
//	LogPrintf("before IoCallDriver %p %p\n",pdx->LowerDeviceObject, Irp));
	status = IoCallDriver(pdx->LowerDeviceObject, Irp);
//	LogPrintf("after  IoCallDriver %p %p\n",pdx->LowerDeviceObject, Irp));
	IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
	return status;
}

void DumpBuffer(struct Buffer *b, unsigned char * buf, int len)
{
#define NB_BYTE 16 /* number of bytes displayed per line */

	char str[NB_BYTE*3 + 1];
	int i,j;

	for (i=0;i<len;i+=NB_BYTE)
	{
		char * p = str;

		for (j=i;j<len&&j<i+NB_BYTE;j++)
		{
			char c;

			*p++ = ' ';
			c = (buf[j] >> 4) & 0xf;
			*p++ = (c<10) ? c+'0' : c-10+'a';
			c = buf[j] & 0xf;
			*p++ = (c<10) ? c+'0' : c-10+'a';
		}
		*p = 0;

		BufferPrintf(b,"    %x:%s\n",i,str);
	}
}

void DumpBufferRaw(struct Buffer *b, unsigned char * buf, int len)
{
	int nlen, nsize ;
	char * nstr;
	unsigned char * databuf;

	if (len < 0)
	{
		/* Nothing to write	*/
		return ;
	}

	// bitmap size is the original data minus its header, plus bitmap header
	if ((len>0xA000) && (len<0xC000)) // the second part of the actual scan, no headers needed
		nlen = len;
	else
		nlen = len - 0x40 + BITMAP_HEADERSIZE;

	nsize = ((nlen + PAGE_SIZE) / PAGE_SIZE) * PAGE_SIZE;

	// this _could_ be trouble: ExAllocatePool did not produce compiler error about types in the C-file
	nstr = (char *)ExAllocatePool(NonPagedPool, nsize);
	if (nstr == NULL)
	{
		myAllocationFailed ++;
		return ;
	}

	/* copy memory block */
	databuf = buf+0x40;
	
	if (len < 0xA000) { //this is the checksum
		RtlMoveMemory(nstr, BitMapHeaderChecksum, BITMAP_HEADERSIZE);
		RtlMoveMemory(nstr+BITMAP_HEADERSIZE, databuf,len-0x40);
	} else if (len < 0xC000) { // this is the second part of a scan
		RtlMoveMemory(nstr, buf,len);
	} else { // first part of the scan
		RtlMoveMemory(nstr, BitMapHeaderScan, BITMAP_HEADERSIZE);
		RtlMoveMemory(nstr+BITMAP_HEADERSIZE, databuf,len-0x40);
	}

	b->str = nstr;
	b->len = nlen;
	b->size = nsize;
}

/*
	Dump a buffer. Its address is either pBuffer or pMdl.
	It dumps for uBufferSize bytes starting at
	pBuffer (or pMDL) + uBufferOffset. This later is optionnal (currently)
*/

void DumpTransferBuffer(struct Buffer *b,
						PUCHAR pBuffer, PMDL pMdl, ULONG uBufferSize,
						BOOLEAN bPrintHeader,ULONG uBufferOffset = 0)
{
	if(bPrintHeader)
	{
		BufferPrintf(b,"  TransferBufferLength = %x\n", uBufferSize);
		BufferPrintf(b,"  TransferBuffer       = %x\n", pBuffer);
		BufferPrintf(b,"  TransferBufferMDL    = %x\n", pMdl);
	}
	else
	{
		if(pBuffer)
		{
			if(pMdl)
			{
				//LogPrintf("??? weird transfer buffer, both MDL and flat specified. Ignoring MDL\n"));
			}
			DumpBuffer(b,pBuffer+uBufferOffset,uBufferSize);
		}
		else if(pMdl)
		{
			PUCHAR pMDLBuf = (PUCHAR)MmGetSystemAddressForMdl(pMdl);
			if(pMDLBuf)
				DumpBuffer(b,pMDLBuf+uBufferOffset,uBufferSize);
			else
			{
				BufferPrintf(b,"XXXXX ERROR: can't map MDL!\n");
			}
		}
		else
		{
			BufferPrintf(b,"\n    no data supplied\n");
		}
	}
}

void DumpTransferBufferRaw(struct Buffer *b,
						PUCHAR pBuffer, PMDL pMdl, ULONG uBufferSize,
						BOOLEAN bPrintHeader,ULONG uBufferOffset = 0)
{
	if(pBuffer)
	{
		if(pMdl)
		{
			//LogPrintf("??? weird transfer buffer, both MDL and flat specified. Ignoring MDL\n"));
		}
		DumpBufferRaw(b,pBuffer+uBufferOffset,uBufferSize);
	}
	else if(pMdl)
	{
		PUCHAR pMDLBuf = (PUCHAR)MmGetSystemAddressForMdl(pMdl);
		if(pMDLBuf)
			DumpBufferRaw(b,pMDLBuf+uBufferOffset,uBufferSize);
		else
		{
			BufferPrintf(b,"XXXXX ERROR: can't map MDL!\n");
		}
	}
	else
	{
		BufferPrintf(b,"\n    no data supplied\n");
	}
}


void DumpGetStatusRequest(struct Buffer *b,
						  struct _URB_CONTROL_GET_STATUS_REQUEST *pGetStatusRequest,
						  BOOLEAN bReturnedFromHCD)
{
	DumpTransferBuffer(b,(PUCHAR)pGetStatusRequest->TransferBuffer, pGetStatusRequest->TransferBufferMDL, pGetStatusRequest->TransferBufferLength, TRUE);
	if(pGetStatusRequest->TransferBufferLength != 1)
		BufferPrintf(b,"  *** error - TransferBufferLength should be 1!\n");
	if(bReturnedFromHCD)
	{
		DumpTransferBuffer(b,(PUCHAR)pGetStatusRequest->TransferBuffer, pGetStatusRequest->TransferBufferMDL, pGetStatusRequest->TransferBufferLength, FALSE);
	}

	BufferPrintf(b,"  Index                = %x\n", pGetStatusRequest->Index);
	
	if(pGetStatusRequest->UrbLink)
	{
		BufferPrintf(b,"---> Linked URB:\n");
		DumpURB(b,pGetStatusRequest->UrbLink, bReturnedFromHCD);
		BufferPrintf(b,"---< Linked URB\n");
	}
}

void DumpFeatureRequest(struct Buffer *b,
						struct _URB_CONTROL_FEATURE_REQUEST *pFeatureRequest,
						BOOLEAN bReadFromDevice, BOOLEAN bReturnedFromHCD)
{
	BufferPrintf(b,"  FeatureSelector = %x\n", pFeatureRequest->FeatureSelector);
	BufferPrintf(b,"  Index           = %x\n", pFeatureRequest->Index);
	if(pFeatureRequest->UrbLink)
	{
		BufferPrintf(b,"---> Linked URB:\n");
		DumpURB(b,pFeatureRequest->UrbLink, bReturnedFromHCD);
		BufferPrintf(b,"---< Linked URB\n");
	}
}

void DumpDescriptorRequest(struct Buffer *b,
						   struct _URB_CONTROL_DESCRIPTOR_REQUEST *pDescriptorRequest,
						   BOOLEAN bReadFromDevice, BOOLEAN bReturnedFromHCD)
{
	DumpTransferBuffer(b,(PUCHAR)pDescriptorRequest->TransferBuffer, pDescriptorRequest->TransferBufferMDL, pDescriptorRequest->TransferBufferLength, TRUE);
	if(((!bReadFromDevice) && (!bReturnedFromHCD)) || (bReadFromDevice && bReturnedFromHCD))
	{
		DumpTransferBuffer(b,(PUCHAR)pDescriptorRequest->TransferBuffer, pDescriptorRequest->TransferBufferMDL, pDescriptorRequest->TransferBufferLength, FALSE);
	}

	BufferPrintf(b,"  Index                = %x\n", pDescriptorRequest->Index);
	BufferPrintf(b,"  DescriptorType       = %x (%s)\n", pDescriptorRequest->DescriptorType,
		pDescriptorRequest->DescriptorType == USB_DEVICE_DESCRIPTOR_TYPE ? "USB_DEVICE_DESCRIPTOR_TYPE" :
		pDescriptorRequest->DescriptorType == USB_CONFIGURATION_DESCRIPTOR_TYPE ? "USB_CONFIGURATION_DESCRIPTOR_TYPE" :
		pDescriptorRequest->DescriptorType == USB_STRING_DESCRIPTOR_TYPE ? "USB_STRING_DESCRIPTOR_TYPE" : "<illegal descriptor type!>");
	BufferPrintf(b,"  LanguageId           = %x\n", pDescriptorRequest->LanguageId);
	
	if(pDescriptorRequest->UrbLink)
	{
		BufferPrintf(b,"---> Linked URB:\n");
		DumpURB(b,pDescriptorRequest->UrbLink, bReturnedFromHCD);
		BufferPrintf(b,"---< Linked URB\n");
	}
}

void DumpVendorOrClassRequest(struct Buffer *b,
							  struct _URB_CONTROL_VENDOR_OR_CLASS_REQUEST *pFunctionClassInterface, BOOLEAN bReturnedFromHCD)
{
	BOOLEAN bReadFromDevice = (BOOLEAN)(pFunctionClassInterface->TransferFlags & USBD_TRANSFER_DIRECTION_IN);
	BufferPrintf(b,"  TransferFlags          = %x (%s, %sUSBD_SHORT_TRANSFER_OK)\n", pFunctionClassInterface->TransferFlags,
		bReadFromDevice ? "USBD_TRANSFER_DIRECTION_IN" : "USBD_TRANSFER_DIRECTION_OUT",
		pFunctionClassInterface->TransferFlags & USBD_SHORT_TRANSFER_OK ? "":"~");

	DumpTransferBuffer(b,(PUCHAR)pFunctionClassInterface->TransferBuffer, pFunctionClassInterface->TransferBufferMDL, pFunctionClassInterface->TransferBufferLength, TRUE);
	if(((!bReadFromDevice) && (!bReturnedFromHCD)) || (bReadFromDevice && bReturnedFromHCD))
	{
		DumpTransferBuffer(b,(PUCHAR)pFunctionClassInterface->TransferBuffer, pFunctionClassInterface->TransferBufferMDL, pFunctionClassInterface->TransferBufferLength, FALSE);
	}

	BufferPrintf(b,"  UrbLink                 = %x\n", pFunctionClassInterface->UrbLink);
	BufferPrintf(b,"  RequestTypeReservedBits = %x\n", pFunctionClassInterface->RequestTypeReservedBits);
	BufferPrintf(b,"  Request                 = %x\n", pFunctionClassInterface->Request);
	BufferPrintf(b,"  Value                   = %x\n", pFunctionClassInterface->Value);
	BufferPrintf(b,"  Index                   = %x\n", pFunctionClassInterface->Index);
	if(pFunctionClassInterface->UrbLink)
	{
		BufferPrintf(b,"---> Linked URB:\n");
		DumpURB(b,pFunctionClassInterface->UrbLink, bReturnedFromHCD);
		BufferPrintf(b,"---< Linked URB\n");
	}
}

void DumpPipeHandle(struct Buffer *b,const char *s,
					USBD_PIPE_HANDLE inPipeHandle)
{
	unsigned char ep;

	// search for the matching endpoint

	if (GetEndpointInfo(inPipeHandle,&ep))
		BufferPrintf(b,"%s = %p [endpoint 0x%x]\n",s,inPipeHandle,ep);
	else
		BufferPrintf(b,"%s = %p\n",s,inPipeHandle);
}

// return value tells whether to start a new file after dumping this buffer
// 0 = don't start, 1 do start
int DumpURB(struct Buffer *b, PURB pUrb, BOOLEAN bReturnedFromHCD)
{
	USHORT wFunction, wLength;
	USBD_STATUS lUsbdStatus;
	int iRetVal=0;

	if (pUrb == NULL)
	{
		BufferPrintf(b,"UsbSnoop - URB == NULL ???\n");
		return 0;
	}

	if (pUrb->UrbHeader.Length < sizeof(pUrb->UrbHeader))
	{
		BufferPrintf(b,"UsbSnoop - incorrect UrbHeader.Length=%d, should be at least %d\n",
			pUrb->UrbHeader.Length,sizeof(pUrb->UrbHeader));
		return 0;
	}

	wFunction = pUrb->UrbHeader.Function;
	wLength = pUrb->UrbHeader.Length;
	lUsbdStatus = pUrb->UrbHeader.Status;

	switch(wFunction)
	{
	case URB_FUNCTION_SELECT_CONFIGURATION:
		{

#define URB_SELECT_CONFIGURATION_SIZE 24

		}
		break; 

	case URB_FUNCTION_SELECT_INTERFACE:
		break;

	case URB_FUNCTION_ABORT_PIPE: // tested [22/11/2001]
		break;

	case URB_FUNCTION_TAKE_FRAME_LENGTH_CONTROL: // untested
		break;

	case URB_FUNCTION_RELEASE_FRAME_LENGTH_CONTROL: // untested
		break;

	case URB_FUNCTION_GET_FRAME_LENGTH: // untested
		break;

	case URB_FUNCTION_SET_FRAME_LENGTH: // untested
		break;

	case URB_FUNCTION_GET_CURRENT_FRAME_NUMBER: // untested
		break;

	case URB_FUNCTION_CONTROL_TRANSFER: // tested [22/11/2001]
		break;

	case URB_FUNCTION_BULK_OR_INTERRUPT_TRANSFER: // tested [22/11/2001]
		{
			struct _URB_BULK_OR_INTERRUPT_TRANSFER *pBulkOrInterruptTransfer = (struct _URB_BULK_OR_INTERRUPT_TRANSFER *) pUrb;

			if(pBulkOrInterruptTransfer->Hdr.Length < sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER ))
			{
				BufferPrintf(b,"!!! Hdr.Length is wrong! (is: %d, should be at least: %d)\n",
					pBulkOrInterruptTransfer->Hdr.Length,
					sizeof(struct _URB_BULK_OR_INTERRUPT_TRANSFER));
				return 0;
			}

			BOOLEAN bReadFromDevice = (BOOLEAN)(pBulkOrInterruptTransfer->TransferFlags & USBD_TRANSFER_DIRECTION_IN);

			// only print inbound traffic and skip event messages
			if(bReadFromDevice && bReturnedFromHCD && (pBulkOrInterruptTransfer->TransferBufferLength > 0x100))  
			{
				DumpTransferBufferRaw(b,(PUCHAR)pBulkOrInterruptTransfer->TransferBuffer, pBulkOrInterruptTransfer->TransferBufferMDL, pBulkOrInterruptTransfer->TransferBufferLength, FALSE);
				myIncomingMsgCount++;
				// A crude way of knowing, when to start a new file: the length of the transfer is around 0xB1C0 (last part of the FP scan)
				// or around 0x8740
				int iLen = pBulkOrInterruptTransfer->TransferBufferLength;
				if (((iLen > 0x8730) && (iLen < 0x8750)) || ((iLen > 0xB1B0) && (iLen < 0xB1D0)))
					iRetVal=1; 
			}

		}
		break;

	case URB_FUNCTION_ISOCH_TRANSFER: // tested [22/11/2001]
		break;

	case URB_FUNCTION_RESET_PIPE:
		break;

	case URB_FUNCTION_GET_DESCRIPTOR_FROM_DEVICE:
		break;

	default:
		 // BufferPrintf(b,"******* non printable URB with function code 0x%x ********\n", wFunction);
		break;
	}	// end of mega switch

	return iRetVal;
}

NTSTATUS MyInternalIOCTLCompletion(IN PDEVICE_OBJECT fido, IN PIRP Irp, IN PVOID inContext)
{
	struct Buffer b;
	int iStartNewFile=0;

	PCONTEXT Context = (PCONTEXT)inContext;
//	DumpContext(Context);

//	if (KeGetCurrentIrql() >= DISPATCH_LEVEL)
//		KeInsertQueueDpc (&gDPC,NULL,NULL);

	// restoring information.

	Context->Stack->CompletionRoutine = Context->CompletionRoutine;
	Context->Stack->Context           = Context->Context;
	Context->Stack->Control           = Context->Control;

	// dumping URB 

	RawBufferInit(&b);
//	BufferPrintf(&b," <<<  URB %d coming back  <<< \n",Context->uSequenceNumber);
	iStartNewFile = DumpURB(&b,Context->pUrb, TRUE);

	RawLogBuffer(&b);
	RawBufferDone(&b);

	if (iStartNewFile == 1)
		LogStartNewFile2();


	// saving some field before freeing the structure
	PVOID OldContext = Context->Context;
	PIO_COMPLETION_ROUTINE OldCompletionRoutine = Context->CompletionRoutine;

	// freeing the allocated structure
	ExFreePool(Context);

	// calling the old CompletionRoutine, if there was one

	if (OldCompletionRoutine != NULL)
		return OldCompletionRoutine(fido,Irp,OldContext);

	return STATUS_SUCCESS;
}

NTSTATUS MyDispatchInternalIOCTL(IN PDEVICE_OBJECT fdo, IN PIRP Irp)
{
//	LogPrintf("UsbSnoop - MyDispatchInternalIOCTL(%p) : fdo=%p, Irp=%p, IRQL=%d\n",
//		MyDispatchInternalIOCTL,fdo,Irp,KeGetCurrentIrql());

//	DumpDeviceObject(fdo);
//	DumpIrp(Irp);

//	if (KeGetCurrentIrql() >= DISPATCH_LEVEL)
//		KeInsertQueueDpc (&gDPC,NULL,NULL);

	// we want to recover OriginalDriverObject which is stored in fido,
	// so we need to recover fido first, by the AttachedDevice field from FDO.

	PDEVICE_OBJECT fido = fdo->AttachedDevice;
//	LogPrintf("  fido = %p\n",fido));

	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
//	LogPrintf("  OriginalDriverObject = %p\n",pdx->OriginalDriverObject));

	// try to print the URB 

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG dwControlCode = stack->Parameters.DeviceIoControl.IoControlCode;
//	LogPrintf("  dwControlCode=%d\n",dwControlCode));
	if (dwControlCode == IOCTL_INTERNAL_USB_SUBMIT_URB)
	{
		struct Buffer b;
		ULONG uSequenceNumber = InterlockedIncrement((PLONG)&pdx->uSequenceNumber);

		RawBufferInit(&b);

	//	BufferPrintf(&b," >>>  URB %d going down  >>> \n", uSequenceNumber);

		// we cannot call ZwWriteFile() cause it must be running at
		// IRQL_PASSIVE_LEVEL (0) and most of the time, MyDispatchInternalIOCTL
		// is running at IRQL_DISPATCH_LEVEL (2).
		// LogPrintf(" KeGetCurrentIrql() = %d\n",KeGetCurrentIrql()));

		PURB pUrb = (PURB) stack->Parameters.Others.Argument1;
		int iStartNewFile = DumpURB(&b,pUrb,FALSE);

		// inspired from the macro code of IoSetCompletionRoutine
		// it just makes a big BSOD

		// ok. in fact, it worked, but that particular case,
		// the first parameter of the IoCompletionRoutine is NULL instead of being fido !!!

		// normally, there should be a call to IoCopyCurrentIrpStackLocationToNext()
		// which might not be there. So, we are surely replacing other callbacks.

		// first, we saved every information we'll modify later

		/*
			Windows 2000 DDK : Note that the context area cannot be pageable
			because the IoCompletion routine can be called at IRQL
			DISPATCH_LEVEL
		*/

		PCONTEXT Context = (PCONTEXT)ExAllocatePool(NonPagedPool,sizeof(CONTEXT));
		if (Context != NULL)
		{
			Context->CompletionRoutine = stack->CompletionRoutine;
			Context->Context           = stack->Context;
			Context->Control           = stack->Control;
			Context->pUrb              = pUrb;
			Context->uSequenceNumber   = uSequenceNumber;
			Context->Stack             = stack;

//			LogPrintf("  Replacing %p/%p/%p by ",
//				stack->CompletionRoutine,stack->Context,stack->Control));
			stack->CompletionRoutine = MyInternalIOCTLCompletion;
			stack->Context = Context;
			stack->Control = SL_INVOKE_ON_SUCCESS | SL_INVOKE_ON_ERROR | SL_INVOKE_ON_CANCEL;

//			LogPrintf("%p/%p/%p\n",stack->CompletionRoutine,stack->Context,stack->Control));
		}
		else
			LogPrintf("  ExAllocatePool failed! Can't redirect CompletionRoutine\n");

		RawLogBuffer(&b);
		RawBufferDone(&b);

		if (iStartNewFile==1)
			LogStartNewFile2();

	}

	return pdx->OriginalDriverObject->MajorFunction[IRP_MJ_INTERNAL_DEVICE_CONTROL](fdo,Irp);
}

/*
	DispatchPower - need a special code, since we have to call PoCallDriver()
	instead of IoCallDriver()
*/

NTSTATUS DispatchPower(IN PDEVICE_OBJECT fido, IN PIRP Irp)
{
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG fcn = stack->MinorFunction;
	
	const char* fcnname[] =
	{
		"IRP_MN_WAIT_WAKE",
		"IRP_MN_POWER_SEQUENCE",
		"IRP_MN_SET_POWER",
		"IRP_MN_QUERY_POWER",
	};
	
	if (fcn == IRP_MN_SET_POWER || fcn == IRP_MN_QUERY_POWER)
	{
		const char* sysstate[] =
		{
			"PowerSystemUnspecified",
			"PowerSystemWorking",
			"PowerSystemSleeping1",
			"PowerSystemSleeping2",
			"PowerSystemSleeping3",
			"PowerSystemHibernate",
			"PowerSystemShutdown",
			"PowerSystemMaximum",
		};
		
		const char* devstate[] =
		{
			"PowerDeviceUnspecified",
			"PowerDeviceD0",
			"PowerDeviceD1",
			"PowerDeviceD2",
			"PowerDeviceD3",
			"PowerDeviceMaximum",
		};
		
		ULONG context = stack->Parameters.Power.SystemContext;
		POWER_STATE_TYPE type = stack->Parameters.Power.Type;
		
	//	LogPrintf("UsbSnoop - IRP_MJ_POWER (%s), SystemContext %x", fcnname[fcn], context);
	//	if (type == SystemPowerState)
	//		LogPrintf(", SystemPowerState = %s\n", sysstate[stack->Parameters.Power.State.SystemState]);
	//	else
	//		LogPrintf(", DevicePowerState = %s\n", devstate[stack->Parameters.Power.State.DeviceState]);
	}
//	else
//		LogPrintf("UsbSnoop - IRP_MJ_POWER (%s)\n", fcnname[fcn]);
		
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
	PoStartNextPowerIrp(Irp);	// must be done while we own the IRP
	NTSTATUS status;
	status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
	if (!NT_SUCCESS(status))
		return CompleteRequest(Irp, status, 0);
	IoSkipCurrentIrpStackLocation(Irp);
	status = PoCallDriver(pdx->LowerDeviceObject, Irp);
	IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
	return status;
}

///////////////////////////////////////////////////////////////////////////////

NTSTATUS DispatchPnp(IN PDEVICE_OBJECT fido, IN PIRP Irp)
{
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG fcn = stack->MinorFunction;
	
	const char * MinorFunctionName = GetIrpPnpMinorFunctionName(fcn);
	if (MinorFunctionName != NULL)
		LogPrintf("UsbSnoop - DispatchPNP(%p) : IRP_MJ_PNP (%s)\n",
			DispatchPnp,MinorFunctionName);
	else
		LogPrintf("UsbSnoop - DispatchPNP(%p) : IRP_MJ_PNP (0x%x)\n",
			DispatchPnp,fcn);
	
	NTSTATUS status;
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
	status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
	if (!NT_SUCCESS(status))
		return CompleteRequest(Irp, status, 0);
	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(pdx->LowerDeviceObject, Irp);
	
	if (fcn == IRP_MN_REMOVE_DEVICE)
	{
		IoReleaseRemoveLockAndWait(&pdx->RemoveLock, Irp);
		RemoveDevice(fido);
	}
	else
		IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
	
	return status;
}

///////////////////////////////////////////////////////////////////////////////

NTSTATUS DispatchWmi(IN PDEVICE_OBJECT fido, IN PIRP Irp)
{							// DispatchWmi
#if DBG
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	ULONG fcn = stack->MinorFunction;
	
	static char* fcnname[] = {
		"IRP_MN_QUERY_ALL_DATA",
		"IRP_MN_QUERY_SINGLE_INSTANCE",
		"IRP_MN_CHANGE_SINGLE_INSTANCE",
		"IRP_MN_CHANGE_SINGLE_ITEM",
		"IRP_MN_ENABLE_EVENTS",
		"IRP_MN_DISABLE_EVENTS",
		"IRP_MN_ENABLE_COLLECTION",
		"IRP_MN_DISABLE_COLLECTION",
		"IRP_MN_REGINFO",
		"IRP_MN_EXECUTE_METHOD",
	};
	
	LogPrintf("UsbSnoop - IRP_MJ_SYSTEM_CONTROL (%s)\n", fcnname[fcn]);
#endif // DBG
	
	NTSTATUS status;
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fido->DeviceExtension;
	status = IoAcquireRemoveLock(&pdx->RemoveLock, Irp);
	if (!NT_SUCCESS(status))
		return CompleteRequest(Irp, status, 0);
	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(pdx->LowerDeviceObject, Irp);
	IoReleaseRemoveLock(&pdx->RemoveLock, Irp);
	return status;
}
