#include "stddcls.h"
#include "MyKdPrint.h"
#include "Driver.h"

struct MyListEntry
{
	LIST_ENTRY	listEntry;
	char *		str;
	int			lenght;
	LARGE_INTEGER currentTime;
};

#pragma LOCKEDCODE
#pragma LOCKEDDATA

#define USE_EVENT
#undef  USE_THREAD

/* USE_EVENT/USE_THREAD : comment (first test using "ECI Telecom USB ADSL Loader")
0/0 : after waiting the 10 specified seconds, a blue screen occur
	with DRIVER_UNLOADED_WITHOUT_CANCELLING_PENDING_OPERATIONS error
0/1 : ok
1/0 : ok
1/1 : ok

  Second test using "ECI Telecom USB ADSL WAN Modem" only

0/0
0/1 : ok
1/0 : ok
1/1 : ok

  Now with both

1/1 : ok. Note : the system wait for the first driver to be unloaded, before
			loading it a second time. That's great!

  Now with both and unlimited messages.

  Windows 98:
  Using "USE_THREAD" do not work and cause BSOD.
*/

KSEMAPHORE mySemaphore;
KSPIN_LOCK mySpinLock;
LIST_ENTRY myListHead;

/*
  myEvent is used to synchronize the destruction of the thread. It is
  first create in a non signaled state. It's signalled when the thread
  is about to call PsTerminateSystemThread().
*/

KEVENT myEvent;

PVOID myThreadObjectPointer = NULL;

LARGE_INTEGER myPerformanceCounter;
LARGE_INTEGER performanceFrequency;

int myMsgTotalCount = 0;
int myMsgValidCount = 0;
int myMsgWrittenCount = 0;
int myBufferWrites = 0;
int myMaxSemaphoreCount = 0;
int	myWaitForsFailed = 0;
int	myNullEntries = 0;
char myCheckChar = 'a';

int myIncomingMsgCount = 0;
int myListInserts = 0;

KIRQL myMaxIrql = 0;
int myAllocationFailed = 0;
int mySemaphoreFailed  = 0;
int myAllocationFailed2 = 0;
int mySemaphoreFailed2 = 0;

char *BitMapHeaderScan;
char *BitMapHeaderChecksum;


// convert the first 4 bit into a char like '0' ...'9', 'A' ... 'F'

const char HexaChar(unsigned int dw)
{
	dw = dw & 0xf;
	if (dw < 10)
		return '0' + dw;
	return 'a' + dw - 10;
}


#define HRES 0x4EBD // these resolutions correspond to the MSFP average
#define VRES 0x4EBD // scan resolution

// Construct a bitmap header (as a char *) for a 8-bit grayscale
// 150dpi image with a given resolution;
char * MakeBitmapHeader(int iLen, int iWidth)
{
	char *hdr = NULL;
	int i=0;int k=0;
	int iFileSize=0;
	int iDataSize=0;
	int iHeaderSize=BITMAP_HEADERSIZE;
	int iHRes = HRES;
	int iVRes = VRES;
	
	hdr = ExAllocatePool(NonPagedPool, BITMAP_HEADERSIZE);
	if (hdr == NULL)
	{
		return NULL;
	}
	
	iDataSize = iLen*iWidth;
	iFileSize = iDataSize+BITMAP_HEADERSIZE;

	// initialize (to be sure)
	for(i=0; i<BITMAP_HEADERSIZE; i++)
		hdr[i]=0;

	// **************** The actual header bytes, all in little-endian format ******************//
	// bitmap identifiers
	hdr[0] = 'B'; hdr[1]='M';

	// filesize
	for(i=0; i<4; i++) {
		hdr[i+2] = iFileSize % 0x100;
		iFileSize = iFileSize / 0x100;
	}
	// four "reserved" bytes starting at 0x0006, left as zero
	// ..
	// header size
	for(i=0; i<4; i++) {
		hdr[i+0xA] = iHeaderSize % 0x100;
		iHeaderSize = iHeaderSize / 0x100;
	}
	// header size without the palettes
	hdr[0xE] = 0x28;
	// image width
	for(i=0; i<4; i++) {
		hdr[i+0x12] = iWidth % 0x100;
		iWidth = iWidth / 0x100;
	}
	// image height
	for(i=0; i<4; i++) {
		hdr[i+0x16] = iLen % 0x100;
		iLen = iLen / 0x100;
	}
	// number of "planes" and the bit-depth of a pixel
	hdr[0x1A]=1; hdr[0x1C]=8;
	// compression is in 0x1E, but we leave it to 0 for no
	// compression.
	// image size
	for(i=0; i<4; i++) {
		hdr[i+0x22] = iDataSize % 0x100;
		iDataSize = iDataSize / 0x100;
	}
	// horizontal and vertical resolution in obscure units: pixels per meter
	for(i=0; i<4; i++) {
		hdr[i+0x26] = iHRes % 0x100;
		iHRes = iHRes / 0x100;
	}
	for(i=0; i<4; i++) {
		hdr[i+0x2A] = iVRes % 0x100;
		iHRes = iVRes / 0x100;
	}
	// number of colors and "important" colors, both 256 from 8-bit grayscale
	hdr[0x2F]=1; hdr[0x33]=1;
	// palette entries, starting from 0x0036
	for (i=0; i<0x100; i++){
		for (k=0; k<3; k++)
			hdr[4*i+k+0x36]=i;
		hdr[4*i+3+0x36]=0;
	}

	return hdr;
}


/*
  MyVsprintfLen : return the length of the string produced by a
  printf().  The length does not include the null terminated
  character. It returns -1 if the format string is invalid (like %C)
*/

int MyVsprintfLen(const char * format, va_list marker)
{
	const char * src;
	unsigned int dw;
	int entier;
	const char * str;
	int diviseur, quotient, reste;

	int r = 0;

	for (src=format;*src!=0;src++)
	{
		if (*src != '%')
		{
			r ++;
			continue;
		}

		switch (* ++src)
		{
		case 'p':
		case 'x':
			dw = va_arg(marker,unsigned int);
			r += 8;
			break;
		case 'd':
			entier = va_arg(marker,int);

			// if 'entier' is negative, make it positive
			if (entier < 0)
			{
				entier = - entier;
				r ++;
			}

			diviseur = 1;
			while (10 * diviseur <= entier)
				diviseur = diviseur * 10;

			do {
				quotient = entier / diviseur;
				reste = entier - diviseur * quotient;
				r ++;

				diviseur = diviseur / 10;
				entier = reste;
			} while (diviseur > 0);

			break;
		case 's':
			str = va_arg(marker,const char *);
			while (*str++!=0)
				r ++;
			break;
		default:
			dw = va_arg(marker,unsigned int);			
			return -1;
			break;
		}
	}

	return r;
}

/*
	MyVsprintf : format a string into 'dst' acording to the format
	in 'format'. Warning : no checking is done on the length of 'dst'.
*/

void MyVsprintf(char *dst, const char * format,va_list marker)
{
	const char * src;
	char cc;
	unsigned int dw;
	int entier;
	const char * str;
	int diviseur, quotient, reste;

	for (src=format;*src!=0;src++)
	{
		if (*src != '%')
		{
			*dst ++ = *src;
			continue;
		}

		switch (* ++src)
		{
		case 'p':
		case 'x':
			dw = va_arg(marker,unsigned int);
			*dst ++ = HexaChar(dw>>28);
			*dst ++ = HexaChar(dw>>24);
			*dst ++ = HexaChar(dw>>20);
			*dst ++ = HexaChar(dw>>16);
			*dst ++ = HexaChar(dw>>12);
			*dst ++ = HexaChar(dw>> 8);
			*dst ++ = HexaChar(dw>> 4);
			*dst ++ = HexaChar(dw    );
			break;
		case 'c':
			cc = va_arg(marker,char);
			*dst ++ = cc;
			break;
		case 'd':
			entier = va_arg(marker,int);

			// if 'entier' is negative, make it positive
			if (entier < 0)
			{
				entier = - entier;
				*dst ++ = '-';
			}


			diviseur = 1;
			while (10 * diviseur <= entier)
				diviseur = diviseur * 10;

			do {
				quotient = entier / diviseur;
				reste = entier - diviseur * quotient;
				*dst ++ = '0' + quotient;

				diviseur = diviseur / 10;
				entier = reste;
			} while (diviseur > 0);

			break;
		case 's':
			str = va_arg(marker,const char *);
			while (*str!=0)
				*dst ++ = *str ++;
			break;
		default:
			dw = va_arg(marker,unsigned int);			
//			DbgPrint("Unhandle type %c\n",*src);
			return ;
			break;
		}

		// int i = va_arg(marker, thetype=int);
	}
	* dst = 0;
}

void MySprintf(char *dst, const char * format, ...)
{
	va_list marker;

	va_start(marker,format);
	MyVsprintf(dst,format,marker);
	va_end(marker);
}

void * OpenTimedLogFile()
{
	OBJECT_ATTRIBUTES objectAttributes;
	UNICODE_STRING unicodeObjectName;
	ANSI_STRING ansiObjectName;
	IO_STATUS_BLOCK ioStatusBlock;
	NTSTATUS status;
	HANDLE hLog;

	LARGE_INTEGER rawSystime;
	TIME_FIELDS sysTime;
	char fileName[100];

	// get the system time (GMT) to mark the log files more easily
	KeQuerySystemTime(&rawSystime);
	RtlTimeToTimeFields(&rawSystime, &sysTime);

	MySprintf(fileName,"\\SystemRoot\\$$$imgs\\Day_%d%d%d%d%d%d_Time_%d%d%d%d%d%d%c.bmp\0",
				(sysTime.Year%100)/10, sysTime.Year%10, (sysTime.Month%100)/10, sysTime.Month%10,
				(sysTime.Day%100)/10,  sysTime.Day%10,  (sysTime.Hour%100)/10, sysTime.Hour%10,
				(sysTime.Minute%100)/10, sysTime.Minute%10, (sysTime.Second%100)/10, sysTime.Second%10,
				myCheckChar);

	// this is to separate different messages arriving at the same second
	// when the seconds differ already, this is redundant
	if (myCheckChar == 'z')
		myCheckChar = 'a';
	else
		myCheckChar++;

//	ZwCreateFile only creates files under SystemRoot, thus the folder containing the images
//  must reside under the Windows-directory 

	RtlInitAnsiString(&ansiObjectName,fileName);

	status = RtlAnsiStringToUnicodeString(&unicodeObjectName,&ansiObjectName,TRUE);
	if (status != STATUS_SUCCESS)
		DbgPrint("RtlAnsiStringToUnicodeString failed, status = 0x%x\n",status);

	InitializeObjectAttributes(&objectAttributes,&unicodeObjectName,0,NULL,NULL);

	// windows 98: using only FILE_APPEND_DATA | SYNCHRONIZE do
	// not work. ZwWriteFile() returns ERROR_INVALID_DATA. So we
	// compute the file position ourselve, but of course, by doing
	// so, the file is truncated each time we open it.

	status = ZwCreateFile(&hLog,FILE_GENERIC_WRITE,
		&objectAttributes,&ioStatusBlock,NULL,FILE_ATTRIBUTE_NORMAL,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		FILE_OVERWRITE_IF,FILE_SYNCHRONOUS_IO_NONALERT,NULL,0);
	if (status != STATUS_SUCCESS)
	{
		DbgPrint("ZwCreateFile failed, status = 0x%x\n",status);
		// to be sure that hLog is NULL
		hLog = (HANDLE)NULL;
	} 

	return hLog;
}

void MyThreadFunction(PVOID arg)
{
	OBJECT_ATTRIBUTES dirObjectAttributes;
	UNICODE_STRING dirUnicodeObjectName;
	ANSI_STRING dirAnsiObjectName;
	IO_STATUS_BLOCK dirIoStatusBlock;
	NTSTATUS status;
	HANDLE hLog;
	HANDLE hOldLog;
	HANDLE hNewLog;
	HANDLE hDir;
	struct MyListEntry * pEntry;
	BOOLEAN bContinue = TRUE;
	LARGE_INTEGER byteOffset;
	ULONG length;

	char buf[200];

	// set the same priority as the lower driver
	KeSetPriorityThread(KeGetCurrentThread(),12);

	// create the log file

	// SystemRoot is something like "C:\WINNT"

	// check that the directory exists, if not, create it
	RtlInitAnsiString(&dirAnsiObjectName,"\\SystemRoot\\$$$imgs\\");

	status = RtlAnsiStringToUnicodeString(&dirUnicodeObjectName,&dirAnsiObjectName,TRUE);
	if (status != STATUS_SUCCESS)
		DbgPrint("RtlAnsiStringToUnicodeString failed, status = 0x%x\n",status);

	InitializeObjectAttributes(&dirObjectAttributes,&dirUnicodeObjectName,0,NULL,NULL);

	// this should just fail, if the directory exists
	status = ZwCreateFile(&hDir,FILE_LIST_DIRECTORY | FILE_TRAVERSE,
		&dirObjectAttributes,&dirIoStatusBlock,NULL,FILE_ATTRIBUTE_NORMAL,
		FILE_SHARE_READ | FILE_SHARE_WRITE,
		FILE_CREATE,FILE_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT,NULL,0);
	if (status != STATUS_SUCCESS)
	{
		DbgPrint("ZwCreateFile failed, status = 0x%x\n",status);
		// to be sure that hLog is NULL
		hDir = (HANDLE)NULL;
	} else {
		ZwClose(hDir); // creating done.
	}

	hLog = OpenTimedLogFile();

	// start at the beginning of the file
	byteOffset.QuadPart = 0;

	// try to remove an entry from the list

	while (bContinue)
	{

		status = KeWaitForSingleObject(&mySemaphore,Executive,KernelMode,FALSE,NULL);
		if (status != STATUS_SUCCESS)
		{
			myWaitForsFailed++;
		}

		pEntry = (struct MyListEntry *)ExInterlockedRemoveHeadList(&myListHead,&mySpinLock);
		if (pEntry == NULL) {
			myNullEntries++;
			continue ;
		}

		myMsgTotalCount++;

		if (pEntry->str == (char *)0xffffffff)
		{
			// a notification to start a new file (if this is not dirty, I don't know what is...)
			hNewLog = OpenTimedLogFile();
			hOldLog = hLog;
			hLog = hNewLog;
			byteOffset.QuadPart = 0;

			if (hOldLog != (HANDLE)NULL)
			{
				status = ZwClose(hOldLog);
				if (status != STATUS_SUCCESS)
					DbgPrint("ZwClose failed : status = 0x%x\n",status);
			}
		} 
		else if ((pEntry->str != NULL) && (pEntry->str != (char *)0xffffffff))
		{
			NTSTATUS status;
			IO_STATUS_BLOCK ioStatusBlock;

			// length = strlen(pEntry->str);

			myMsgValidCount++;

			if (hLog != NULL)
			{
				length = pEntry->lenght;
				status = ZwWriteFile(hLog,NULL,NULL,NULL,&ioStatusBlock,
					(PVOID)pEntry->str,length,&byteOffset,NULL);
				if (status != STATUS_SUCCESS)
					DbgPrint("ZwWriteFile(str) failed : status = 0x%x\n",status);
				byteOffset.QuadPart += length;

				myMsgWrittenCount++;
			}

			ExFreePool(pEntry->str);
		}
		else 
		{
			// if we receive an entry with a NULL string, this is the signal to
			// stop this thread
			bContinue = FALSE;
		}

		ExFreePool(pEntry);
	}

	// close the file if necessary
	if (hLog != (HANDLE)NULL)
	{
		status = ZwClose(hLog);
		if (status != STATUS_SUCCESS)
			DbgPrint("ZwClose failed : status = 0x%x\n",status);
	}

	DbgPrint("Thread terminating ...\n");

#ifdef USE_EVENT
	KeSetEvent(&myEvent,0,FALSE);
#endif

	DbgPrint("Thread is dying\n");
	PsTerminateSystemThread(STATUS_SUCCESS);
}

void MyThreadStart()
{
	HANDLE hThread;
	NTSTATUS status;

	DbgPrint("MyThreadStart()\n");

	KeInitializeSemaphore(&mySemaphore,0,MAX_SEMAPHORE);

	// initialize a double-linked list
	KeInitializeSpinLock(&mySpinLock);
	InitializeListHead(&myListHead);

	KeInitializeEvent(&myEvent, NotificationEvent, FALSE);

	myPerformanceCounter = KeQueryPerformanceCounter(&performanceFrequency);

	/*
		We create a kernel thread.
		From Windows 2000 DDK:
		Callers of this routine must be running at IRQL PASSIVE_LEVEL.
	*/

	status = PsCreateSystemThread(&hThread,THREAD_ALL_ACCESS,NULL,NULL,NULL,
		MyThreadFunction,NULL);
	if (status != STATUS_SUCCESS)
		DbgPrint("PsCreateSystemThread=0x%x\n",status);

	// the following code is compiled only in .sys file for
	// Windows 2000/XP platform because this code make the .sys
	// file unloadable under Windows 98

#ifdef WIN2K
	// we keep an "object pointer" on the thread, for waiting on it later
	ObReferenceObjectByHandle(hThread,THREAD_ALL_ACCESS,NULL,KernelMode,
		&myThreadObjectPointer,NULL);
#endif

	status = ZwClose(hThread);
	if (status != STATUS_SUCCESS)
		DbgPrint("ZwClose=0x%x\n",status);

	DbgPrint("MyThreadStart() : OK\n");

	BitMapHeaderScan = MakeBitmapHeader(SCAN_HEIGHT, SCAN_WIDTH);
	BitMapHeaderChecksum = MakeBitmapHeader(CHECKSUM_HEIGHT, SCAN_WIDTH);

}

void MyThreadStop()
{
	// allocate a new list entry
	struct MyListEntry * pEntry;
	NTSTATUS status;

	DbgPrint("MyThreadStop()\n");
	
	pEntry = (struct MyListEntry *)ExAllocatePool(NonPagedPool,
		sizeof(struct MyListEntry));
	if (pEntry == NULL)
	{
		DbgPrint("Failed to allocate pEntry. Can't stop the thread!\n");
		return ;
	}

	pEntry->str = NULL;

	// add the new entry to the tail of the list, this is used to tell
	// the thread to end its infinite loop

	ExInterlockedInsertTailList(&myListHead,&pEntry->listEntry,&mySpinLock);
	KeReleaseSemaphore(&mySemaphore,0,1,FALSE);

#ifdef USE_EVENT
	// wait for the thread to be destroy
	status = KeWaitForSingleObject(&myEvent,Executive,KernelMode,FALSE,NULL);
	if (status != STATUS_SUCCESS)
		DbgPrint("KeWaitForSingleEvent=%x\n",status);

	DbgPrint("retour de KeWaitForSingleEvent\n");
#endif

	// windows xp: without this code, a BSOD occur when the driver is unloaded, because
	// of DRIVER_UNLOADER_WITH_PENDING_OPERATIONS (or similar)

#ifdef WIN2K
	status = KeWaitForSingleObject(myThreadObjectPointer,Executive,KernelMode,FALSE,NULL);
	if (status != STATUS_SUCCESS)
		DbgPrint("KeWaitForSingleThread=%x\n",status);

	DbgPrint("retour de KeWaitForSingleThread\n");

	ObDereferenceObject(myThreadObjectPointer);
#endif

	DbgPrint("MyThreadStop() : OK\n");
}

/*
	InternalPrintStr : send a string to the log file (it is first sent to
	the thread, that will log to the file)

	Returned value is 1 (success, in this case the string will be freed later)
	or 0 (failure)
*/

int InternalPrintStr(struct Buffer * b)
{
	struct MyListEntry * pEntry;
	LONG semaphoreCount;
	KIRQL currentIrql;

	// check the semaphore counter
	semaphoreCount = KeReadStateSemaphore(&mySemaphore);
	if (!(semaphoreCount < MAX_SEMAPHORE))
	{
		mySemaphoreFailed ++;
		return 0;
	}
		
	// allocate a new list entry
	pEntry = (struct MyListEntry *)ExAllocatePool(NonPagedPool,sizeof(struct MyListEntry));
	if (pEntry == NULL)
	{
		myAllocationFailed ++;
		return 0;
	}

	// copy the string in the newly created entry
	pEntry->str = b->str;
	pEntry->lenght = b->len;
	//KeQuerySystemTime(&pEntry->currentTime);

	/*
		Windows 2000 DDK: Callers of KeQueryPerformanceCounter can be running
		at any IRQL.
	*/
	pEntry->currentTime = KeQueryPerformanceCounter(NULL);

	// add the new entry to the tail of the list
	ExInterlockedInsertTailList(&myListHead,&pEntry->listEntry,&mySpinLock);
	myListInserts++;

	semaphoreCount = KeReleaseSemaphore(&mySemaphore,0,1,FALSE);

	if (semaphoreCount > 0 && (semaphoreCount%100)==0)
		DbgPrint("semaphoreCount=%d\n",semaphoreCount);

	if (myMaxSemaphoreCount < semaphoreCount)
		myMaxSemaphoreCount = semaphoreCount;

	currentIrql = KeGetCurrentIrql();
	if (myMaxIrql < currentIrql)
		myMaxIrql = currentIrql;

	return 1;
}

/* the LogXXX API implementation */

void LogInit()
{
	MyThreadStart();
}

void LogDone()
{
	MyThreadStop();
}

void LogStartNewFile()
{
// ************ First stop the thread without deallocating the list struct ********
}

void LogStartNewFile2()
{
	struct MyListEntry * pEntry;
	LONG semaphoreCount;
	KIRQL currentIrql;

	// check the semaphore counter
	semaphoreCount = KeReadStateSemaphore(&mySemaphore);
	if (!(semaphoreCount < MAX_SEMAPHORE))
	{
		mySemaphoreFailed2++;
		return;
	}
		
	// allocate a new list entry
	pEntry = (struct MyListEntry *)ExAllocatePool(NonPagedPool,sizeof(struct MyListEntry));
	if (pEntry == NULL)
	{
		myAllocationFailed2++;
		return;
	}

	// use a special marker (risky, I know...)
	pEntry->str = (char *)0xffffffff;
	//KeQuerySystemTime(&pEntry->currentTime);

	pEntry->currentTime = KeQueryPerformanceCounter(NULL);

	// add the new entry to the tail of the list
	ExInterlockedInsertTailList(&myListHead,&pEntry->listEntry,&mySpinLock);

	semaphoreCount = KeReleaseSemaphore(&mySemaphore,0,1,FALSE);

	if (semaphoreCount > 0 && (semaphoreCount%100)==0)
		DbgPrint("semaphoreCount=%d\n",semaphoreCount);

	if (myMaxSemaphoreCount < semaphoreCount)
		myMaxSemaphoreCount = semaphoreCount;

	currentIrql = KeGetCurrentIrql();
	if (myMaxIrql < currentIrql)
		myMaxIrql = currentIrql;
}

void LogPrintf(const char * format, ...)
{
	struct Buffer b;
	va_list marker;

	va_start(marker,format);

	BufferInit(&b);
	BufferVprintf(&b,format,marker);
	LogBuffer(&b);
	BufferDone(&b);

	va_end(marker); 
}

void LogBuffer(struct Buffer * b)
{
	if (b->str == NULL)
		return;

	// windows98: we limit the number of character sent to the debugger
	// since this can cause BSOD.

#if 0
	DbgPrint("LogBuffer(str=%p, len=%d)\n",b->str,b->len);

	if (b->len > 100)
		b->len = 100;

	b->str[b->len] = 0;

	DbgPrint(b->str);
#endif

	if (InternalPrintStr(b))
	{
		/*
			If InternalPrintStr succeds, that means the buffer if no longer
			our property. So we reset the field in 'b'
		*/

		b->str  = NULL;
		b->len  = 0;
		b->size = 0;
	}
}

void RawLogBuffer(struct Buffer * b)
{
	if (b->str == NULL)
		return;

	// windows98: we limit the number of character sent to the debugger
	// since this can cause BSOD.

#if 0
	DbgPrint("LogBuffer(str=%p, len=%d)\n",b->str,b->len);

	if (b->len > 100)
		b->len = 100;

	b->str[b->len] = 0;

	DbgPrint(b->str);
#endif

	if (InternalPrintStr(b))
	{
		/*
			If InternalPrintStr succeds, that means the buffer if no longer
			our property. So we reset the field in 'b'
		*/

		b->str  = NULL;
		b->len  = 0;
		b->size = 0;
	}
}
